// ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "3dGfx.cc", created on 13/06/99.
 * descr : 3d gfx functions & normal gfx funcs
 *
 * Ray_Caster() func is written on the base of same-called function by
 * Andre' LaMothe published in book "Tricks of the Game-Programming"
 * by Sams Publishing.
 *
 * Ground texturing funcs are mine. I spent 3 hours thinking (away from
 * computer) about them, and ... you can see.
 */
// ----------------------------------------------------------------------------------

#include <math.h>
#include "z11.h"
#include "WorldMap.h"
#include "3dGfx.h"
#include "Player.h"

FILEINFO(Gfx3d);

int windowHeight;
int windowWidth;

float *tan_table,*inv_tan_table,*y_step,*x_step,*cos_table,*inv_cos_table,*inv_sin_table;
unsigned char * lightDatas;
int * lightTables;

// this is comment from Worlock.c original file... it's not very up-to-date ...
// This is the heart of the system.  it casts out 320 rays and builds the
// 3-D image from their intersections with the walls.  It was derived from
// the previous version used in "RAY.C", however, it has been extremely
// optimized for speed by the use of many more lookup tables and int
// point math

void Draw3dGround(int x, int y, float ang);
extern int debugArg;
// ---------------------------------------------------------------------------------
// assembly interface
void doHDraw() asm("_doHDraw");
void doVDraw() asm("_doVDraw");
void doSpriteVDraw() asm("_doSpriteVDraw"); // ELF does not prepend '_'

int asmdh_sw asm("sw"),
    asmd_n asm("n"),
    asmdh_scale_step asm("scale_step"),
    asmdh_scale_index asm("scale_index");
unsigned char * asmd_text asm("text"),
              * asmd_light asm("light"),
              * asmd_vram asm("vram");
int asmdv_x asm("x"),
    asmdv_y asm("y"),
    asmdv_vx asm("vx"),
    asmdv_vy asm("vy");
// ---------------------------------------------------------------------------------
// objects met by ray. If object meet more than 100 objects - they will be ignored
int nRayObjects;
GameObject * rayObjects[100];
inline void addObject(GameObject * obj) {
  if(!obj) return;
  int n=nRayObjects;
  GameObject ** p = rayObjects;
  while(n-->0) if(*p++==obj) break;
  if(n<0 && nRayObjects<sizeofT(rayObjects))
    rayObjects[nRayObjects++]=obj;
}
// ---------------------------------------------------------------------------------
// cmp func for qsort()
int cmpObjects(const void * o1, const void * o2) {
  if((*(GameObject**)o1)->getDist()<(*(GameObject**)o2)->getDist()) return 1;
  else return -1;
}
// ---------------------------------------------------------------------------------
void Ray_Caster(int x,int y,float view_angle) {
int
cell_x,       // the current cell that the ray is in
cell_y,
casting=2,    // tracks the progress of the X and Y component of the ray
x_bound,      // the next vertical and horizontal intersection point
y_bound,
next_y_cell,  // used to figure out the quadrant of the ray
next_x_cell,

x_delta,      // the amount needed to move to get to the next cell
y_delta,      // position
xb_save,
yb_save,
xi_save,      // used to save exact x and y intersection points
yi_save;

bool xray,       // tracks the progress of a ray looking for Y interesctions
     yray;       // tracks the progress of a ray looking for X interesctions

int
     dist_x,  // the distance of the x and y ray intersections from
     dist_y;  // the viewpoint
float xi,     // used to track the x and y intersections
      yi;

// S E C T I O N  1 /////////////////////////////////////////////////////////v
Draw3dGround(x,y,view_angle);


// initialization

// compute starting angle from player.  Field of view is 60 degrees, so
// subtract half of that current view angle

view_angle-=30;
if (view_angle < 0)   view_angle+=360;
if (view_angle > 360) view_angle-=360;

float vad = (float)60 / windowWidth;

// loop through all 320 rays

for (int ray=windowWidth-1; ray>=0; ray--) {
    int vaf = fixUp(view_angle);
    if (view_angle >= 0 && view_angle < 180) {
       y_bound = cellSize + (y & cellMaskUp);
       y_delta = cellSize;
       xi = inv_tan_table[vaf] * (y_bound - y) + x;
       next_y_cell = 0;
    }
    else {
       y_bound = y & cellMaskUp;
       y_delta = -cellSize;
       xi = inv_tan_table[vaf] * (y_bound - y) + x;
       next_y_cell = -1;
    }

    if (view_angle < (90) || view_angle >= (270)) {
       x_bound = cellSize + (x & cellMaskUp);
       x_delta = cellSize;
       yi = tan_table[vaf] * (x_bound - x) + y;
       next_x_cell = 0;
    }
    else {
       x_bound = x & cellMaskUp;
       x_delta = -cellSize;
       yi = tan_table[vaf] * (x_bound - x) + y;
       next_x_cell = -1;
    }

    casting       = 2;                // two rays to cast simultaneously
    xray=yray     = false;            // reset intersection flags
    Cell * xhit, * yhit;
    int xp = next_x_cell ? Cell::left : Cell::right,
        yp = next_y_cell ? Cell::up   : Cell::down; // ? ? ?
    nRayObjects=0;
    while(casting) {
       if (!xray) {
            cell_x = ( (x_bound+next_x_cell) >> cellFP );
            cell_y = (int)yi;
            cell_y>>=cellFP;

            xhit=map.cell(cell_x,cell_y);
            addObject(xhit->obj);
            // object can be hit in naighbour fields too ...
            addObject(map.cell(cell_x-1,cell_y-1)->obj);
            addObject(map.cell(cell_x,cell_y-1)->obj);
            addObject(map.cell(cell_x-1,cell_y)->obj);
            addObject(map.cell(cell_x+1,cell_y+1)->obj);
            addObject(map.cell(cell_x,cell_y+1)->obj);
            addObject(map.cell(cell_x+1,cell_y)->obj);
            addObject(map.cell(cell_x+1,cell_y-1)->obj);
            addObject(map.cell(cell_x-1,cell_y+1)->obj);
            if (xhit->filled) {
               dist_x  = (int)((yi - y) * inv_sin_table[vaf]);
               yi_save = (int)yi;
               xb_save = x_bound;
               xray = true;
               casting--;
            }
            else {
               yi += y_step[vaf];
               x_bound += x_delta;
            } // end else
       }
       if (!yray) {
            cell_x = (int)xi;
            cell_x>>=cellFP;
            cell_y = ( (y_bound+next_y_cell) >> cellFP );

            yhit = map.cell(cell_x,cell_y);
            addObject(yhit->obj);
            // object can be hit in naighbour fields too ...
            addObject(map.cell(cell_x-1,cell_y-1)->obj);
            addObject(map.cell(cell_x,cell_y-1)->obj);
            addObject(map.cell(cell_x-1,cell_y)->obj);
            addObject(map.cell(cell_x+1,cell_y+1)->obj);
            addObject(map.cell(cell_x,cell_y+1)->obj);
            addObject(map.cell(cell_x+1,cell_y)->obj);
            addObject(map.cell(cell_x+1,cell_y-1)->obj);
            addObject(map.cell(cell_x-1,cell_y+1)->obj);
            if (yhit->filled) {
               dist_y  = (int)((xi - x) * inv_cos_table[vaf]);
               xi_save = (int)xi;
               yb_save = y_bound;
               yray = true;
               casting--;
            } // end if a hit
            else {
               xi += x_step[vaf];
               y_bound += y_delta;
            } // end else
       } // end if y ray has intersected
    } // end while casting

    view_angle+=vad;
    if (view_angle>=360) view_angle-=360;

    int dist,p,save;
    Cell * hit;
    int lightDist;
    if (dist_x < dist_y) {
       dist = dist_x;
       lightDist = dist;
       hit = xhit;
       p=xp;
       save = yi_save; // !
    }
    else {
       dist = dist_y;
       lightDist = dist + cellSize/2; // to make y walls darker ...
       hit = yhit; p=yp;
       save = xi_save; // !
    }

    if(dist<12*cellSize) hit->known=true;

    WallBitmap * bmp = hit->wall[p];
    //assert(bmp);
    if(!bmp) continue;
    int scale = (int)((cos_table[fixUp((float)ray*60/windowWidth)]*windowHeight)/(200*dist));
    if(scale>0) { // could be 0 only for *very* small windows
      save = (bmp->h*(save%cellSize))>>cellFP;
      bmp->x = ray;
      bmp->y = (windowHeight - scale) >> 1;
      bmp->RenderSliver(scale,save,lightDist);
    }

    if(nRayObjects) {
      // ---------------- little bit math ------------------------
      // ray: y=a*x+b
      // object = (xp,yp)
      // distance: d=abs(b+a*xp-yp)/sqrt(a*a+1)

      float a=tan_table[vaf];
      float b=player->y-a*player->x;
      float s2a=sqrt(a*a+1); // maybe put this into table ?

      GameObject ** p = rayObjects;
      for(int i=0; i<nRayObjects; i++) {
        GameObject * obj = rayObjects[i];
        if(obj->getDist()<dist &&
           obj->setColumn(fabs(b+a*obj->x-obj->y)/s2a,ray)) *p++=obj;
      }
      nRayObjects=p-rayObjects;
      if(nRayObjects>1) {
        qsort(rayObjects,nRayObjects,sizeof(rayObjects[0]),cmpObjects);
        p=rayObjects;
        while(nRayObjects-->0) (*p++)->draw();
      }
      else if(nRayObjects==1) {
        (*rayObjects)->getDist();
        (*rayObjects)->draw();
      }
    }
  } // end for ray

} // end Ray_Caster
// ---------------------------------------------------------------------------------
bool GameObject::setColumn(float d, int ray) {
  if(d>size) return false;
  if(setColumnFrame!=frame) {
    setColumnFrame=frame;
    minDist=d;
  }
  if(d>minDist) column=(int)((getBitmap()->h*(size-d))/(size*2));
  else {
    column=(int)((getBitmap()->h*(size+d))/(size*2));
    minDist=d;
  }
  getBitmap()->x=ray;
  return true;
}
// ---------------------------------------------------------------------------------
void Draw3dGround(int x, int y, float ang) {
  const float pi = 3.141592654;
  ang*=pi/180; // -> rad
  const int fbase = 1<<16;
  const int cos30 =(int)(cos(ang+pi/6)*fbase),
            cosm30=(int)(cos(ang-pi/6)*fbase),
            sin30 =(int)(sin(ang+pi/6)*fbase),
            sinm30=(int)(sin(ang-pi/6)*fbase);
  int n = windowHeight/2;
  //const float mv = (float)fbase/windowWidth;
  x<<=16; y<<=16;
  int scaleFloor=(int)(vScale*map.lightFloor);
  int scaleCeil=(int)(vScale*map.lightCeil);
  for(int i=0; i<n; i++) {
    int w=(windowHeight-(i<<1));
    int p=(windowWidth<<cellFP)/w;
    int xl=x+cos30*p,  yl=y+sin30*p,
        xr=x+cosm30*p, yr=y+sinm30*p;
    int pvx=(int)((xr-xl)/windowWidth),
        pvy=(int)((yr-yl)/windowWidth);

    asmdv_x=xl;   asmdv_y=yl;
    asmdv_vx=pvx; asmdv_vy=pvy;
    asmd_n=windowWidth;
    asmd_vram = doubleBuffer.line[windowHeight-i-1];
    asmd_text = map.floorBmp.buffer;
    asmd_light = lightDatas+lightTables[((scaleFloor/w)<?(maxLightTab-1))>?1];
    doHDraw();

    asmdv_x=xl;   asmdv_y=yl;
    asmdv_vx=pvx; asmdv_vy=pvy;
    asmd_n=windowWidth;
    asmd_vram = doubleBuffer.line[i];
    asmd_text = map.ceilBmp.buffer;
    asmd_light = lightDatas+lightTables[((scaleCeil/w)<?(maxLightTab-1))>?1];
    doHDraw();
  }
}
// ---------------------------------------------------------------------------------
void GameObject::draw() {
  assert(angleOk==frame);
  assert(distOk==frame);

  int scale = (int)((vScale*windowHeight)/(200*cachedDist));
  cachedBitmap->y = (windowHeight - scale) >> 1; //!!!!
  cachedBitmap->RenderSliver(scale,column,cachedDist);
}

// ----------------------------------------------------------------------------------
// draws vertical sliver
void WallBitmap::RenderSliver(int scale, int column, int dist)
{
  asmdh_scale_step=(w << 8) / scale;
  dist-=overboard-1;
  dist>>=(cellFP-6);
  if(dist<=0) dist=1;
  if(dist>=maxLightTab) dist=maxLightTab-1;
  if(y < 0) {
    asmdh_scale_index = -y * asmdh_scale_step;
    y=0;
    scale = windowHeight;
  }
  else {
    asmdh_scale_index = 0;
  }
  asmd_text=line[column];
  // it's only called column, since WallBitmap::load() roates bitmap
  // to make this work properly
  asmd_light=lightDatas+lightTables[dist];
  // compute offset of sprite in video double buffer
  asmd_vram = doubleBuffer.line[y]+x;
  asmdh_sw = doubleBuffer.w;
  asmd_n=scale;
  doVDraw();
}
// ----------------------------------------------------------------------------------
// draws vertical sliver for sprites - colro zero is transparent
void SpriteBitmap::RenderSliver(int scale, int column, int dist)
{
  asmdh_scale_step=(w << 8) / scale;
  dist-=overboard-1;
  dist>>=(cellFP-6);
  if(dist<=0) dist=1;
  if(dist>=maxLightTab) dist=maxLightTab-1;
  if(y < 0) {
    asmdh_scale_index = -y * asmdh_scale_step;
    y=0;
    scale = windowHeight;
  }
  else {
    asmdh_scale_index = 0;
  }
  asmd_text=line[column];
  // it's only called column, since WallBitmap::load() roates bitmap
  // to make this work properly
  asmd_light=lightDatas+lightTables[dist];
  // compute offset of sprite in video double buffer
  asmd_vram = doubleBuffer.line[y]+x;
  asmdh_sw = doubleBuffer.w;
  asmd_n=scale;
  doSpriteVDraw();
}





