// ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "Objects.cc", created on 25/06/99.
 * descr : all Objects: static and moving. before 25/06/99 there was no other object
 * then player described by 3 variables ...
 *
 */
// ----------------------------------------------------------------------------------
#include <math.h>

#include "z11.h"
#include "WorldMap.h"
#include "3dGfx.h"
#include "Player.h"
#include "NPC.h"
#include "Gun.h"
#include "FileRead.h"

FILEINFO(Objects);

// ----------------------------------------------------------------------------------
float tgAng[90],sinAng[360],cosAng[360];
static Player p;
Player * player = &p; // 4now...
extern SpriteBitmap * jedi;
// ----------------------------------------------------------------------------------
void MovingObject::handleEvent(Event * e) {
  if(e->className()=="EvMove") {
    if(me) delete me;
    me=(EvMove*)e;
  }
  else GameObject::handleEvent(e);
}
// ----------------------------------------------------------------------------------
GameObject::GameObject() :
  cachedBitmap(0),
  angBitmaps(false),
  angleOk(-1),
  distOk(-1),
  setColumnFrame(-1),
  x(0),
  y(0),
  size(cellSize/3),
  angle(frand(360))
  { }
// ----------------------------------------------------------------------------------
MovingObject::MovingObject() :
  vx(0),
  vy(0),
  maxV(4.0*cellSize/ups),
  maxA((4.0*cellSize/ups)/(0.6*ups)), // 0.6 sec to full speed
  omega(0),
  maxOmega(100.0/ups),
  maxAOmega((140.0/ups)/(0.4*ups)), // 0.4 sec to full omega
  me(0)
  { }
// ----------------------------------------------------------------------------------
void normalize(float & angle) {
  while(angle<0) angle+=360;
  while(angle>360) angle-=360;
}
// ----------------------------------------------------------------------------------
// time is in ups/sec (ups==200, now)
// if you don't feel well in kinematics - better don't look at that
float speed=1.0,ospeed=1.0; // could be changed from console
void MovingObject::move(int time) {
  if(!me) return;
  float an=angle,dvx=0,dvy=0;
  // sorry, didn't think about this. know that ugly.
  float maxV=this->maxV*speed;
  float maxA=this->maxA*speed;
  float maxOmega=this->maxOmega*ospeed;
  float maxAOmega=this->maxAOmega*ospeed;
  // desired V
  // know, that walking back and left/right should be slower, but I (and not only I)
  // like playing w/o that slowmotion more
  if(me->dir!=EvMove::dirNone) {
    switch(me->dir) {
      case EvMove::dirRight:     an-=90;  break;
      case EvMove::dirLeft:      an+=90;  break;
      case EvMove::dirBackward:  an+=180; break;
      default: break;
    }
    normalize(an);
    dvx=maxV*cosAng[(int)an];
    dvy=maxV*sinAng[(int)an];
  }

  float ax=(dvx-vx)/time,ay=(dvy-vy)/time;
  if(ax*ax+ay*ay>maxA*maxA) {
    if(ay!=0) {
      float d=ax/ay, sy=ay>0?1:-1;
      ay=sqrt(maxA*maxA/(d*d+1))*sy;
      ax=d*ay;
    }
    else ax=maxA*(ax>0?1:-1);
  }
  vx+=ax*time; vy+=ay*time;

  if(vx*vx+vy*vy>maxV*maxV) {
    if(vy!=0) {
      float d=vx/vy, sy=vy>0?1:-1;;
      vy=sqrt(maxV*maxV/(d*d+1))*sy;
      vx=d*vy;
    }
    else vx=maxV*(vx>0?1:-1);
  }
  // omega ...
  switch(me->turn) {
    case EvMove::turnNone:
      if(omega>0) {
        if(omega>maxAOmega*time) omega-=maxAOmega*time;
        else omega=0;
      }
      else if(omega<0) {
        if(-omega>maxAOmega*time) omega+=maxAOmega*time;
        else omega=0;
      }
      break;
    case EvMove::turnRight:
      omega-=maxAOmega*time;
      if(-omega>maxOmega) omega=-maxOmega;
      break;
    case EvMove::turnLeft:
      omega+=maxAOmega*time;
      if(omega>maxOmega) omega=maxOmega;
      break;
    default: halt(); // ???
  }

  // move
  if(canMove(x+vx*time,y+vy*time)) { x+=vx*time; y+=vy*time; }
  else if(canMove(x+vx*time,y)) { x+=vx*time; vy=0; } // hit the wall
  else if(canMove(x,y+vy*time)) { y+=vy*time; vx=0; } // as below
  else { vx=0; vy=0; } // bang!
  // and turn
  angle+=omega*time;
  normalize(angle);
}
// ----------------------------------------------------------------------------------
int biSearch(float * t, float v, int l, int r) {
  if(l==r) return r;
  if(l-r==1) {
    if(fabs(t[l]-v)<fabs(t[r]-v)) return l;
    else return r;
  }
  int m=(l+r)/2;
  if(t[m]<v) return biSearch(t,v,m,r);
  else return biSearch(t,v,l,m);
}
// ----------------------------------------------------------------------------------
bool isIn(float v, const char * z) {
  istrstream s(z);
  char ld,rd,c;
  float l,r;
  s >> ld >> l >> c >> r >> rd;
  if(ld=='<' && v<l) return false;
  else if(ld=='(' && v<=l) return false;
  if(rd=='>' && v>r) return false;
  else if(rd==')' && v>=r) return false;
  return true;
}
// ----------------------------------------------------------------------------------
SpriteBitmap * GameObject::getBitmap() {
  const float r2d=180/M_PI;
  if(/*!angBitmaps ||*/ angleOk==frame) return cachedBitmap;
  angleOk=frame;
  return cachedBitmap=getBitmapAng(atan2(y-player->y,x-player->x)*r2d);
}
// ----------------------------------------------------------------------------------
Block <void*> guns;
Weapon * gunNo(int i) {
  if(i<guns.size()) return (Weapon*) guns[i];
  else return 0;
}
// ----------------------------------------------------------------------------------
void Player::think(int) {
 // ...
}
// ----------------------------------------------------------------------------------
bool canGo(int x, int y, int size) {
  return !map.cell((x-size)/cellSize,(y-size)/cellSize)->filled &&
         !map.cell((x-size)/cellSize,(y+size)/cellSize)->filled &&
         !map.cell((x+size)/cellSize,(y-size)/cellSize)->filled &&
         !map.cell((x+size)/cellSize,(y+size)/cellSize)->filled &&
         !map.cell((x)/cellSize,(y+size)/cellSize)->filled &&
         !map.cell((x+size)/cellSize,(y)/cellSize)->filled &&
         !map.cell((x)/cellSize,(y-size)/cellSize)->filled &&
         !map.cell((x-size)/cellSize,(y)/cellSize)->filled &&
         !map.cell((x)/cellSize,(y)/cellSize)->filled ;
}
// ----------------------------------------------------------------------------------
// enemy isn't going anywhare
void NPC::think(int) {
 // ...
}
// ----------------------------------------------------------------------------------
SpriteBitmap * NPC::getBitmapAng(float) {
  return jedi;
}
// ----------------------------------------------------------------------------------
float GameObject::getDist() {
  if(distOk!=frame) {
    distOk=frame;
    cachedDist=sqrt((x-player->x)*(x-player->x)+(y-player->y)*(y-player->y));
  }
  return cachedDist;
}
// ----------------------------------------------------------------------------------
bool GameObject::teleport(float tx, float ty) {
  if(!canMove(tx,ty)) return false;
  Cell * target = map.cell((int)tx>>cellFP,(int)ty>>cellFP);
  if(target->obj) return false;
  Cell * prev=map.cell((int)x>>cellFP,(int)y>>cellFP);
  if(prev->obj==this) prev->obj=0;
  target->obj=this;
  x=tx; y=ty;
  return true;
}
// ----------------------------------------------------------------------------------
/*
syntax of gun description :
cnu == currently not used
gun [name] {
  [startUp true] // player has it, when he starts, def false cnu
  [ammo shells] // name of ammo, cnu
  shot 232145654 // shot sequence, one char => one frame
  normal 1 // this describes normal sequance when walking
           // only 1st frames is displayed 4 now
  gfx gfx-data/guns/shotgun/%s.pcx // %s is relpaced with char from shot
  [shotTime 1.5] // in secs, def 1.0
  [width 320]
  [height 200] // window dimensions for which gfx has been made - as you may
  // guess - there are used for scaling
}
*/
void Weapon::read(FileReader & s) {
  s.goIn();
    name=s.sectionName();
    nx=s.getInt("width",nx);
    ny=s.getInt("height",ny);
    if( (shotTime=(int)(s.getFloat("shotTime",(float)shotTime/ups)*ups)) == 0 )
      s.error("shot time must not be zero!");
    if((gfxPath=s.getString("gfx",gfxPath)) == "")
      s.error("gfx path is required in gun description");
    if((shotS=s.getString("shot",shotS)) == "")
      s.error("shot sequance is required in gun description");
    if((normalS=s.getString("normal",normalS)) == "")
      s.error("normal sequance is required in gun description");
    XStr p = xprintf(gfxPath,"%c"); // to change %s => %c
    try {
      for(int i=0; i<shotS.length(); i++)
        if(shotFrames.add(xprintf(p,shotS[i]))!=i) halt();
      for(int i=0; i<normalS.length(); i++)
        if(normalFrames.add(xprintf(p,normalS[i]))!=i) halt();
    } catch(EFile & e) {
      s.error(e.format());
    }
    XStr deltas=s.getString("deltas","");
    for(int i=0; i<deltas.length(); i++) {
      WeaponBitmap * b = shotFrames.find(xprintf(p,deltas[i]));
      if(b) b->x=s.getInt(XStr("delta")+deltas[i],0);
      b = normalFrames.find(xprintf(p,deltas[i]));
      if(b) b->x=s.getInt(XStr("delta")+deltas[i],0);
    }
  s.goOut();
}
// ----------------------------------------------------------------------------------
bool Weapon::shoot() {
  if(shooting) return false;
  shooting=true;
  shotStart = timer();
  return true;
}
// ----------------------------------------------------------------------------------
void Weapon::updateScreen(MetaBmp & buffer) {
  test();
  if(timer()-shotStart>shotTime) shooting=false;
  if(shooting) 
    shotFrames[(timer()-shotStart)*shotFrames.size()/shotTime]
                                                   ->put(buffer,nx,ny);
  else normalFrames[0]->put(buffer,nx,ny);
}
// ----------------------------------------------------------------------------------
void readGuns(XStr fname) {
  FileReader s;
  if(!s.open(fname)) haltx("cant find file "+fname);

  try {
    while(s.nextClassWord()) {
      assert(s.topLevel());
      XStr word=downcase(s.classWord);
      if(word=="gun") {
        Weapon * x = new Weapon();
        x->read(s);
        guns.add(x);
      }
      else s.error(xprintf("unreconaized keyword '%s'",word.chars()));
    }
  }
  catch(EParseError & e) {
    throw EDead(e.xmsg); // 4 now
  }

  s.close();
}
// ----------------------------------------------------------------------------------
Bitmap * AnimBitmapBase::find(XStr name) {
  int i=datas.index(Binder(name));
  if(i==-1) return 0;
  else return (Bitmap *) datas[i].userData;
}
// ----------------------------------------------------------------------------------
int AnimBitmapBase::add(XStr fileName) {
  int i=datas.index(Binder(fileName));
  if(i==-1) {
    Bitmap * b = createNew();
    b->load(fileName);
    return indexes.add(datas.add(Binder(fileName,b)));
  }
  else
    return indexes.add(i);
}
// ----------------------------------------------------------------------------------
AnimBitmapBase::~AnimBitmapBase() {
  for(int i=0; i<datas.size(); i++)
    if(datas[i].userData)
      delete (Bitmap*) datas[i].userData;
}
// ----------------------------------------------------------------------------------
void Player::updateGfx() {
  if(gunNo(activeGun))
    gunNo(activeGun)->updateScreen(doubleBuffer);
}
// ----------------------------------------------------------------------------------
Player::Player() {
  activeGun=0;
}
// ----------------------------------------------------------------------------------
Bitmap * AnimBitmapBase::operator [] (int i) {
  test();
  assert(i>=0);
  assert(i<indexes.size());
  assert(indexes[i]>=0);
  assert(indexes[i]<datas.size());
  // to get error here not in Block.h
  return (Bitmap*)datas[indexes[i]].userData;
}
// ----------------------------------------------------------------------------------




