 // ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "WorldMap.cc", created on 16/06/99.
 * descr : Map & Cell classes
 * and random map generation. wonna thnx Misto for his idea how to do it. know, that
 * idea is simple, but I was thinking about such "great" and "simple" method as
 * neuron nets...
 * I've tested Life-based generation alogoritms, but dungens looked like ... don't
 * know, it was mad. Maybe I'll use it in some rooms.
 * I wonder how Thomas Biskup did it ...
 *
 */
// ----------------------------------------------------------------------------------

#include <ctype.h>
#include "z11.h"
#include "WorldMap.h"
#include "3dGfx.h"
#include "NPC.h"
#include "FileRead.h"

FILEINFO(WorldMap);


extern XStr gfxBlank;
// ----------------------------------------------------------------------------------
class SeedCell {
  public:
    float p;
    XStr name;
    bool filled,door;
    WallBitmap * wall[4];
    SeedCell();
};
// ----------------------------------------------------------------------------------
class CellSelector {
  public:
    Block <SeedCell*> infos;
    SeedCell * rnd();
    bool contains(SeedCell * cell) { return infos.index(cell)!=-1; }
    void add(SeedCell * item) { infos.add(item); }
};
// ----------------------------------------------------------------------------------
SeedCell * CellSelector::rnd() {
  float s=0;
  for(int i=0; i<infos.size(); i++)
    s+=infos[i]->p;
  float r=s*rand()/RAND_MAX;
  s=0;
  int i;
  for(i=0; i<infos.size(); i++) {
    s+=infos[i]->p;
    if(s>=r) break;
  }
  assert(i<infos.size());
  return infos[i]; // used as "color"
}
// ----------------------------------------------------------------------------------
class MapGenerator {
    void find(CellSelector & type, int & x, int & y, bool seeAround = false);
  public:
    CellSelector solid, floor, standAlone;
    Map * map;
    SeedCell * cells[64][64];
    SeedCell * & cell(int x, int y) {
      return cells[x][y];
    }
    void rect(int x1, int y1, int x2, int y2, SeedCell * color);
    void line(int x1, int y1, int x2, int y2, SeedCell * color);
    void generate();
    MapGenerator(Map * p) : map(p) { }
};
// ----------------------------------------------------------------------------------
void MapGenerator::rect(int x1, int y1, int x2, int y2, SeedCell * color) {
  int xl=x1<?x2, yl=y1<?y2, xr=x1>?x2, yr=y1>?y2;
  for(int i=xl; i<=xr; i++)
    for(int j=yl; j<=yr; j++)
      cell(i,j)=color;
}
// ----------------------------------------------------------------------------------
void MapGenerator::line(int x1, int y1, int x2, int y2, SeedCell * color) {
  if(x1==x2 && y1==y2) { 
    if(x1>=0 && y1>=0 && x1<map->w && y1<map->h) cell(x1,y1>>16)=color;
    return; 
  }
  if(abs(x1-x2)>abs(y1-y2)) {
    if(x1>x2) { swap(x1,x2); swap(y1,y2); } 
    int d=((y2-y1)<<16)/(x2-x1);
    y1<<=16;
    int mh=map->h<<16;
    while(x1<=x2) {
      if(x1>=0 && y1>=0 && x1<map->w && y1<mh) cell(x1,y1>>16)=color;
      y1+=d;
      x1++;
    }
  }
  else {
    if(y1>y2) { swap(x1,x2); swap(y1,y2); } 
    int d=((x2-x1)<<16)/(y2-y1);
    x1<<=16;
    int mw=map->w<<16;
    while(y1<=y2) {
      if(x1>=0 && y1>=0 && x1<mw && y1<map->h) cell(x1>>16,y1)=color;
      x1+=d;
      y1++;
    }
  }
}
// ----------------------------------------------------------------------------------
void MapGenerator::find(CellSelector & type, int & x, int & y, bool seeAround) {
  const int timeOut=100000;
  int i;
  for(i=0; i<timeOut; i++) {
    x=rand()%(map->w-2)+1;
    y=rand()%(map->h-2)+1;
    if(seeAround) {
      if(
        type.contains(cell(x,y)) &&
        type.contains(cell(x-1,y)) &&
        type.contains(cell(x,y-1)) &&
        type.contains(cell(x+1,y)) &&
        type.contains(cell(x,y+1)) &&
        type.contains(cell(x-1,y+1)) &&
        type.contains(cell(x+1,y-1)) &&
        type.contains(cell(x-1,y-1)) &&
        type.contains(cell(x+1,y+1))
      ) break;
    }
    else {
      if(type.contains(cell(x,y))) break;
    }
  }
  assert(i<timeOut);
}
// ----------------------------------------------------------------------------------
// rolls number using standard RPG convention : 3d6 - means roll three times
// dice with six wall and add what you'll get, 1d8+10 roll one dice with 8 walls,
// and add ten to result. d6 - roll once
int roll(XStr x) {
  istrstream s(x);
  int n,m=0,add=0;
  s >> n;
  if(!s) { s.clear(); n=1; } // "d8" for ex, without number of dices - assume 1
  char c;
  s >> c;
  if(tolower(c)!='d' && tolower(c)!='k') haltx("bad input format : %s",x.chars());
  // that 'k' sits deep in my mind ... dice in Polish starts with 'k'
  s >> m;
  if(!m) haltx("bad input format [2] : %s",x.chars());
  if(s) s >> add;
  for(int i=0; i<n; i++)
    add+=rand()%m+1; // +1 since you can't get 0 rolling dice ...
  return add;
}
// ----------------------------------------------------------------------------------
inline int rx(int numb) { return rand()%numb; }
// ----------------------------------------------------------------------------------
void MapGenerator::generate() {
  rect(0,0,map->w-1,map->h-1,solid.rnd());
  for(int n=roll("5d8"); n>0; n--)
    rect(rx(map->w),rx(map->h),rx(map->w),rx(map->h),solid.rnd());
  // generate rooms coordinates
  int rooms=roll("2d7+2");
  int roomX[rooms], roomY[rooms]; // gnu rules!
  for(int i=0; i<rooms; i++) {
    int x,y;
    find(solid,x,y,true);
    roomX[i]=x; roomY[i]=y;
  }
  // link & draw rooms
  for(int i=0; i<rooms; i++) {
    if(i<rooms-1) {
      SeedCell * c=floor.rnd();
      if(rand()&1) {
        line(roomX[i],roomY[i],roomX[i+1],roomY[i],c);
        line(roomX[i+1],roomY[i+1],roomX[i+1],roomY[i],c);
      }
      else {
        line(roomX[i],roomY[i],roomX[i],roomY[i+1],c);
        line(roomX[i+1],roomY[i+1],roomX[i],roomY[i+1],c);
      }
    }
    int h=rx(map->h/20)+map->h/40;
    int w=rx(map->w/20)+map->w/40;
    rect(roomX[i]-w>?1,         roomY[i]-h>?1,
         roomX[i]+w<?(map->w-2),roomY[i]+h<?(map->h-2),
         floor.rnd());
  }
  for(int n=rand()%(map->w/15)+2; n>0; n--) {
    int x,y;
    find(floor,x,y,true);
    cell(x,y)=standAlone.rnd();
  }
  // that's all 4 now, convert this bitmap into map ;)
  for(int i=0; i<map->w; i++)
    for(int j=0; j<map->h; j++) {
      map->cells[i][j]=new Cell(cell(i,j));
    }
  find(floor,map->startX,map->startY);
}
// ----------------------------------------------------------------------------------
Map map;
FileReader s;
// ----------------------------------------------------------------------------------
void Map::readMap() {
  s.goIn();
    if(s.emptySection()) s.error("map must not be empty section!");
    name=s.sectionName();
    if(name=="") name=s.getFileName();

    h = s.getInt("height",64);
    w = s.getInt("width",64);
    if(w>64 || h>64) 
      s.error("sorry but maps bigger then 64x64 not supported yet.");

    picture.setSize(w,h);
    
    ceilBmp.load(s.getString("ceil",gfxBlank));
    floorBmp.load(s.getString("floor",gfxBlank));
    ceilBmp.resize(floorSize,floorSize);
    floorBmp.resize(floorSize,floorSize);
    
    //ceilBmp.lock();
    //floorBmp.lock();
    
    lightFloor=s.getFloat("floorShadow",2);
    lightCeil= s.getFloat("ceilShadow",2);
  s.goOut();
}
// ----------------------------------------------------------------------------------
void Map::read(XStr fname) {
  walls.reset();
  w=h=0; // for error detection

  if(!s.open(fname)) haltx("cant find file "+fname);

  try {
    while(s.nextClassWord()) {
      assert(s.topLevel());
      XStr word=downcase(s.classWord);
      if(word=="wall") makeWall();
      else if(word=="cell") makeCell();
      else if(word=="map") readMap();
      else s.error(xprintf("unreconaized keyword '%s'",word.chars()));
    }
    if(w==0 || h==0) s.error("map width or height == 0 - probably 'map' keyword not found");
  }
  catch(EParseError & e) {
    throw EDead(e.xmsg); // 4 now
  }

  s.close();
}
// ----------------------------------------------------------------------------------
Cell * cellStd;
void Map::makeCell() {
  s.goIn();
    SeedCell * r = new SeedCell();
    r->name = s.sectionName();
    if(s.emptySection()) s.error("cell cannot be empty section");

    r->filled = s.getBool("filled",true);
    r->door   = s.getBool("door",false);
    if(s.goIn("wall")) {
        if(s.emptySection()) {
          WallBitmap * p = (WallBitmap*)s.findBind(s.sectionName(),walls);
          for(int i=0; i<sizeofT(r->wall); i++) r->wall[i]=p;
        }
        else {
          r->wall[Cell::left] =(WallBitmap*)s.getBind("left", "def",walls);
          r->wall[Cell::right]=(WallBitmap*)s.getBind("right","def",walls);
          r->wall[Cell::up]   =(WallBitmap*)s.getBind("up",   "def",walls);
          r->wall[Cell::down] =(WallBitmap*)s.getBind("down", "def",walls);
        }
      s.goOut();
    }

    r->p=s.getFloat("prob",1);
    if(s.getBool("standAlone",false)) gen->standAlone.add(new SeedCell(*r));
    if(s.getBool("wallPart",r->filled&&!r->door)) gen->solid.add(new SeedCell(*r));
    if(!r->filled) gen->floor.add(new SeedCell(*r));
    if(r->filled && !cellStd) cellStd=new Cell(r);
    delete r;
  s.goOut();
}
// ----------------------------------------------------------------------------------
void Map::makeWall() {
  WallBitmap * r = new WallBitmap();
  s.goIn();
    if((r->name = s.sectionName()) == "") s.error("could not create unnamed wall");
    if(s.emptySection()) s.error("wall cannot be empty section");

    XStr fname=s.getString("file","");
    if(fname=="") s.error("file name must be be given!");
    try {
      r->load(fname);
    } catch(EFile & e) {
      s.error(e.format()); // to get extendent info about line no etc
    }
  s.goOut();
  walls.add(Binder(r->name,r));
}
// ----------------------------------------------------------------------------------
Cell * Map::cell(int x, int y) {
  if(x<0 || x>=w || y<0 || y>=h)
      // halt();
      return cellStd; // empty but non-transparent ...
  else
     return cells[x][y];
}
// ----------------------------------------------------------------------------------
SeedCell::SeedCell() : filled(true), door(false) {
  memset(wall,0,sizeof(wall));
}
// ----------------------------------------------------------------------------------
Cell::Cell(SeedCell * p) : filled(p->filled), door(p->door), known(false) {
  for(int i=0; i<sizeofT(wall); i++)
    wall[i]=p->wall[i];
  obj=0;
  setId();
}
// ----------------------------------------------------------------------------------
Map::Map() {
  memset(cells,0,sizeof(cells));
  gen=new MapGenerator(this);
}
// ----------------------------------------------------------------------------------
Map::~Map() {
  deleteCells();
  delete gen;
}
// ----------------------------------------------------------------------------------
void Map::deleteCells() {
  for(int i=0; i<w; i++)
    for(int j=0; j<h; j++)
      if(cells[i][j]) delete cells[i][j];
}
// ----------------------------------------------------------------------------------
void Map::generate() {
  deleteCells();
  gen->generate();
  // put some objets on the map ...
  for(int i=0; i<20; i++) {
    NPC * e = new NPC();
    while(!e->teleport(frand(w*cellSize),frand(h*cellSize))) { }
  }
}
// ----------------------------------------------------------------------------------
