// ----------------------------------------------------------------------------------
/* [z-11 project]
 * (c) 1999 Copyright Malekith
 * <malekith@topnet.pl> <http://malekith.topnet.pl>
 * For info about copying see file copying.z11.
 *
 * file "init.cc", created on 15/06/99.
 * descr : initialization and exit routines.
 */
// ----------------------------------------------------------------------------------

#include <sys/param.h>
#include <unistd.h>
#include <math.h>
#include <fstream.h>

#include "z11.h"
#include "Gfx.h"
#include "Input.h"
#include "3dGfx.h"
#include "WorldMap.h"
#include "Player.h"
#include "Term.h"

FILEINFO(init);

// ----------------------------------------------------------------------------------
void readGuns(XStr fname);
void my2ArgErrorHandler(const char * m1, const char * m2) __attribute__ ((noreturn));
// ----------------------------------------------------------------------------------
void growWindow(int dx, int dy);
void printTopLine();
extern char * copyrightTxt, * helpTxt;

extern float tgAng[90], sinAng[360], cosAng[360];

Bmp doubleBuffer, background;

// gfx mode
int xRes=320, yRes=200;

// light
float wstart=1,kstart=7,wm=0.9,km=1.05,darkSide=1;

char version[256];

bool quitFast=false, ansiTerminal=true;

SpriteBitmap * jedi; // tmp bitmap ...
extern int debugArg;  // for some strange debug actions

// args, set by main()
extern int argc;
extern char ** argv;

// directories
XStr txtData, gfxData, lngData, dataPath, gfxBlank;

#define z11printf(args...) ({ printf(args); printTopLine(); })
// ----------------------------------------------------------------------------------
void printTopLine() {
  static bool topLineOk=false;
  const int scrW = 80;
 // let's suppose ...
  static char topLine[scrW+1] = "Z-11 Operating System v" VERSION;
  if(!topLineOk) {
    XStr t = replicate(" ",scrW);
    t.at((int)(scrW-strlen(topLine))/2,(int)strlen(topLine))=topLine;
    assert(t.length()==scrW);
    strcpy(topLine,t);
    topLineOk=true;
  }
# ifdef DJGPP
// not everyone have ansi.sys installed under doz
    int x,y;
    ScreenGetCursor(&x,&y);
    ScreenPutString(topLine,(BLUE<<4)|YELLOW,0,0);
    ScreenSetCursor(x,y);
# else
// on the other hand -- linux console is ansi compatile, so i don't have
// to use curses ...
    if(ansiTerminal) 
      printf("\e[s\e[0;0H\e[0;44m\e[1;33m%s\e[0m\e[u",topLine);
# endif
}
// ----------------------------------------------------------------------------------
// nice, isn't it ?
void initTextConsole() {
  printf(replicate("\n",100));
  printTopLine();
}
// ----------------------------------------------------------------------------------
void buildLightTable(ColourPalette & pal) {
  z11printf("buildLightTable(): ");

  lightTables = new int[maxLightTab+1];
  const int maxTabs=256;
  lightDatas = new byte[maxTabs*256];

  byte * p = lightDatas;

  float w=wstart;
  float k=kstart;
  int n=0;

  for(int i=0; i<maxTabs; i++) {
    w*=wm;
    k*=km;
    RGB * src = &pal[0];
    int nn=256;
    while(nn-->0) {
      RGB c = *src++;
      c.r = (byte)((c.r*w)<?maxByte);
      c.g = (byte)((c.g*w)<?maxByte);
      c.b = (byte)((c.b*w)<?maxByte);
      *p++=pal.nearestTo(c);
    }
    
    if((i&0x3)==0) z11printf("."); // to mark progress

    for(int z=1; z<k; z++) {
      if(n>=maxLightTab) break;
      lightTables[n++]=i*256;
    }
    if(n>=maxLightTab) break;
  }
  assert(n==maxLightTab);
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void removeSinCosTables() {
  delete [] tan_table;
  delete [] inv_tan_table;
  delete [] y_step;
  delete [] x_step;
  delete [] cos_table;
  delete [] inv_cos_table;
  delete [] inv_sin_table;
  delete [] lightTables;
  delete [] lightDatas;
}
// ----------------------------------------------------------------------------------
// this function builds all the look up tables for the system
void buildSinCosTables() {
  z11printf("buildSinCosTables(): ");

  // allocate memory for all look up tables tangent tables equivalent to slopes
  tan_table     = new float[fixUp(361)]; inv_tan_table = new float[fixUp(361)];
  // step tables used to find next intersections, equivalent to slopes times width and height of cell
  y_step        = new float[fixUp(361)]; x_step        = new float[fixUp(361)];
  // cos table used to fix view distortion caused by caused by radial projection
  cos_table     = new float[fixUp(361)];
  // 1/cos and 1/sin tables used to compute distance of intersection very quickly
  inv_cos_table = new float[fixUp(361)]; inv_sin_table = new float[fixUp(361)];

  // create tables, sit back for a sec!
  for (int ang=0; ang<=fixUp(360); ang++) {
      float rad_angle = (float)((3.272e-4) + ang * 2*3.141592654/fixUp(360));
      tan_table[ang]     = (float)tan(rad_angle);
      inv_tan_table[ang] = (float)(1/tan_table[ang]);
      // tangent has the incorrect signs in all quadrants except 1, so
      // manually fix the signs of each quadrant since the tangent is
      // equivalent to the slope of a line and if the tangent is wrong
      // then the ray that is case will be wrong
      if (ang>=fixUp(0) && ang<fixUp(180))
         y_step[ang]        = (float)(fabs(tan_table[ang]  * cellSize));
      else
         y_step[ang]        = (float)(-fabs(tan_table[ang] * cellSize));
      if (ang>=fixUp(90) && ang<fixUp(270))
         x_step[ang]        = (float)(-fabs(inv_tan_table[ang] * cellSize));
      else
         x_step[ang]        = (float)(fabs(inv_tan_table[ang]  * cellSize));

      inv_cos_table[ang] = (float)(1/cos(rad_angle));
      inv_sin_table[ang] = (float)(1/sin(rad_angle));
  }

  // create view filter table.  There is a cosine wave modulated on top of
  // the view distance as a side effect of casting from a int point.
  // to cancel this effect out, we multiple by the inverse of the cosine
  // and the result is the proper scale.  Without this we would see a
  // fishbowl effect, which might be desired in some cases?
  for (int ang=-fixUp(30); ang<=fixUp(30); ang++) {
      float rad_angle = (float)((3.272e-4) + ang * 2*3.141592654/fixUp(360));
      cos_table[ang+fixUp(30)] = (float)(vScale/cos(rad_angle));
  } // end for
  // those tavles are mine, those above are from original Worlock
  const float pi=3.141592654;
  for(int i=0; i<90; i++) tgAng[i]=tan(i*pi/180);
  for(int i=0; i<360; i++) { sinAng[i]=sin(i*pi/180); cosAng[i]=cos(i*pi/180); }
  z11printf("ok\n");
} // end Build_Tables
// ----------------------------------------------------------------------------------
char defines[] =
#   ifdef linux
    "linux "
#   endif
#   ifdef svgalib
    "svgalib "
#   endif
#   ifdef DJGPP
    "DJGPP "
#   endif
#   ifdef NDEBUG
    "NDEBUG "
#   endif
#   ifdef DSX
    "DSX "
#   endif
#   ifdef AllegroTimer
    "AllegroTimer "
#   endif
#   ifdef BlockCheck
    "QuickBlock "
#   endif
#   ifdef gfxCheck
    "gfxCheck "
#   endif
  "VERSION = " VERSION;

void parseCommandLine() {
# define stopArg(args...) \
  ( { \
    z11printf("\ncommand line error:\n"); \
    z11printf(args); \
    z11printf("\ncmdline: `%s'\n",inputLine); \
    exit(1);\
  } )
  z11printf("parseCommandLine(): ");
  // args
  ostrstream os;
  {
    ifstream f(txtData+"config");
    while(f) {
      XStr word;
      f >> word;
      if(!f) break;
      os << word << " ";
    }
  }
  for(int i=1; i<argc; i++) os << argv[i] << " ";
  os << '\0';
  char * inputLine=os.str();
  istrstream args(inputLine);
  while(args) {
    XStr p;
    args >> p;
    p.downcase();
    if(p=="") break;
    if(p=="-m" || p=="--mode") {
      args >> xRes >> yRes;
      if(!args) stopArg("2 numbers (resolution) must follow driver...");
    }

#   define light(v,sn,ln) \
      else if(p==sn || p==ln) \
      { args >> v; if(!args) \
      stopArg( ln " (" sn ") : must be followed by a real number" ); }

    light(wstart,"-ms","--multipler-start")
    light(kstart,"-ps","--pixel-start")
    light(km,"-pv","--pixel-value")
    light(wm,"-mv","--multipler-value")
    light(darkSide,"-ds","--dark-side")

#   undef light

    else if(p=="-h"||p=="--help"||p[0]=='?') {
      z11printf(helpTxt, wstart, kstart, wm, km, darkSide);
      exit(0);
    }
    else if(p=="-d" || p=="--debug-arg") {
      args >> debugArg;
      if(!args) stopArg("-d must be followed by a debugArg value (number)");
    }
    else if(p=="-v" || p=="--version") {
      z11printf("\n\n%s\n\n",version);
      exit(0);
    }
    else if(p=="-df" || p=="--defines") {
      XStr s="\n#defines used for compilation :\n";
      z11printf(s+defines);
      exit(0);
    }
    else stopArg("option not reconized: %s",p.chars());
  }
  delete inputLine;
  z11printf("ok\n");
# undef stopArg
}
// ----------------------------------------------------------------------------------
void loadBackGround() {
  z11printf("loadBackGround(): ");
  Bmp pcx;
  pcx.load(gfxData+"textures/backgd.pcx");
  Bmp::remapTo=Bmp::lastImagePal;
  Bmp::remapTo[0]=((RGB){0,0,0}); // first color have to be black (for some
  // obscure reason ;)
  background.setSize(xRes,yRes);
  for(int x=0; x<background.w; x+=pcx.w)
    for(int y=0; y<background.h; y+=pcx.h)
      pcx.put(background,x,y);
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void initWindow() {
  z11printf("initWindow(): ");
  windowHeight=screen.h;
  windowWidth=screen.w;
  growWindow(0,0);
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void initPlayer() {
  z11printf("initPlayer(): ");
  player->x=map.startX*cellSize+cellSize/2;
  player->y=map.startY*cellSize+cellSize/2;
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void initMap() {
  z11printf("initMap(): srand() ");
  srand(time(0)); // seed random number genrerator
  z11printf("ok\ninitMap(): Map::read() ");
  map.read(txtData+"map");
  z11printf("ok\ninitMap(): Map::generate() ");
  map.generate();
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
// little bit limited ...
void initNPC() {
  z11printf("initNPC(): ");
  jedi=new SpriteBitmap();
  jedi->load(gfxData+"jedi.pcx");
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void initGfx() {
  z11printf("initGfx(): setting screen mode ");
  screen.setRes(xRes,yRes);
  screen.palette=blackPalette;
  screen.clear();
  kbd->turnOn();
  z11printf("ok\n"); // -> you'll see it after exiting game
}
// ----------------------------------------------------------------------------------
void initKeyboard() {
  z11printf("initKeyboard: reading file "+txtData+"keyboard");
  kbd->init(txtData+"keyboard");
  kbd->moveTarget=player;
  kbd->commandTarget=term;
  z11printf(" ok\n");
}
// ----------------------------------------------------------------------------------
void initItems() {
  z11printf("initItems(): readGuns(): ");
  readGuns(txtData+"guns");
  z11printf("ok\n");
}
// ----------------------------------------------------------------------------------
void cleanUp() {
  removeSinCosTables();
  if(dbg) fclose(dbg);
  kbd->kill();
  if(quitFast) { screen.kill(); return; }
  term->executeFile(txtData+"exit.z11");
  delete term;
  screen.palette.fadeOut(500);
  screen.kill();
  // viet - very inteligent ending texts
  char * s[] = {
    "don't go away!",
    "i'm so sad... you left...",
    "see ya soon!",
    "we're saying 'goodbye' before we said 'helo'",
    "why are you leaving me?",
    "there's not always some tomorrow."
  };
  z11printf("z-11 version %s, %s\n%s\n",
            VERSION,
            copyrightTxt,
            s[irand(sizeof(s)/sizeof(s[0]))]
            );
}
// ----------------------------------------------------------------------------------
void setDirs() {
  z11printf("setDirs(): ");
  char path[MAXPATHLEN];
  if(!realpath(argv[0],path)) 
    haltErrno("couldn't extract path name from %s",argv[0]);

# ifdef DJGPP
  char drive[MAXDRIVE],dir[MAXDIR];
  fnsplit(path,drive,dir,0,0);
  dataPath=XStr(drive)+dir;
# else
  dataPath=XStr(path).before("/",-1)+"/";
# endif

  txtData=dataPath+"txt-data/";
  gfxData=dataPath+"gfx-data/";
  lngData=dataPath+"texts-en/";
  gfxBlank=gfxData+"textures/blank.pcx";
  
  z11printf("main dir: %s ok\n",dataPath.chars());
}
// ----------------------------------------------------------------------------------
void init() {
  // calls vga_init -> gives up root rights -> call it w/o unneeded noise ...
  screen.init(); 
  
  lib_error_handler = my2ArgErrorHandler;
  initTextConsole();
  sprintf(version,"z-11 v%s, build: %d (%s)",VERSION,buildNo,buildTime);
  
  setbuf(stdout,0);
  z11printf("z-11 v%s, %s\n",VERSION,copyrightTxt);
  z11printf("init():\n");
  
  setDirs();
  parseCommandLine();
  loadBackGround();
  initMap();
  initItems();
  initNPC();
  initPlayer();

  buildSinCosTables();
  buildLightTable(Bmp::remapTo);

  initKeyboard();
  initGfx();
  initWindow();
  term->init();

  // display the background
  background.put(0,0);
  screen.palette.fadeIn(500,Bmp::remapTo);
  term->executeFile(txtData+"startup.z11");
}
// ----------------------------------------------------------------------------------




